/***************************    classify.c    **************************

Purpose: Classify a given set of test. 
         The classification result is written to the prediction file.

Notes: None.
Programmer: Ginny Mak
Location: Montreal, Canada.
Last Updated: April 6, 2000
**/

#include <stdio.h>
#include <time.h>
#include <math.h>
#include "classify.h"
#include "initialize.h"

/*************************   Declared Functions   ************************/

static double dotProduct( FeaturePtr *x, int sizeX, FeaturePtr *y, int sizeY );
static double power( double x, int n );
static double wtDotProduct( double *w, int sizeX, FeaturePtr *y, int sizeY );
static void freeMemory( void );


/****************************  Private Functions   **********************

/*FN*********************************************************************
  
static double dotProduct( FeaturePtr *x, int sizeX, FeaturePtr *y, int sizeY )

Returns: the dot product of two given vectors

Purpose: Calculate the dot product of two given data vectors.

Note: This function does not change the passed arguments.
**/

static double dotProduct( FeaturePtr *x, int sizeX, FeaturePtr *y, int sizeY )
{
  int num1, num2, a1, a2;
  int p1 = 0; int p2 = 0;
  double dot = 0;

  if( sizeX == 0 || sizeY == 0 )
    return 0;
  num1 = sizeX; num2 = sizeY;
  while( p1 < num1 && p2 < num2 ) {
    a1 = x[p1]->id;
    a2 = y[p2]->id;
    if( a1==a2 ) {
      dot += (x[p1]->value) * (y[p2]->value);
      p1++; p2++;
    }
    else if ( a1 > a2 )
      p2++;
    else
      p1++;
  }
  return dot;
}


/*FN*************************************************************************

double wtDotProduct( double *w, int sizeX, FeaturePtr *y, int sizeY )

Returns: double -- the dot product between the weight vector  and an example.

Purpose: To speed up classifying by just using the weight vector 
         when using linear kernel.

Notes: This function only applies to linear kernel.
**/

static double wtDotProduct( double *w, int sizeX, FeaturePtr *y, int sizeY )
{
  int num1, num2, a2;
  int p1 = 1; int p2 = 0;
  double dot = 0;

  if( sizeX == 0 || sizeY == 0 )
    return 0;
  num1 = sizeX; num2 = sizeY;
  while( p1 <= num1 && p2 < num2 ) {
    a2 = y[p2]->id;
    if( p1==a2 ) {
      dot += (w[p1]) * (y[p2]->value);
      p1++; p2++;
    }
    else if ( p1 > a2 )
      p2++;
    else
      p1++;
  }
  return dot;
}


/*FN***********************************************************************
 
static double power( double x, int n )

Returns: double -- the value of x raised to the power of n

Purpose: To calculate x raised to the power of n.

Notes: x can be positive or negative. n >= 0
**/

static double power( double x, int n )
{
  int i;
  double p;

  p = 1.0;
  for ( i = 1; i <= n; i++ )
    p = p*x;
  return p;
}

/*FN**************************************************************************

static void freeMemory( void )

Returns: Nothing.

Purpose: Free memory allocated in the program.

Notes: None.
**/

static void freeMemory( void )
{
  int i;

  free( lambda );
  free( svNonZeroFeature );
  free( nonZeroFeature );
  free( target );
  free( weight );
  free( output );
}


/***************************   Public functions   ***************************/

/*FN*************************************************************************
 
int writeResult( FILE *out )

Returns: int -- Fail(0) on errors, Succeed(1) otherwise

Purpose: To call classification functions to classify the test data
         according to the kernel type and then write the results 
	 to the prediction file.

Note: Results are written to the prediction file.
**/

int writeResult( FILE *out )
{
  int result;
	  
  if ( kernelType == 0 )
    result = classifyLinear( out );
   
  else if ( kernelType == 1 )
    result = classifyPoly( out ) ;
	      
  else if ( kernelType == 2 )
    result = classifyRbf( out );

  if (result)
    return 1;                       
  return 0;
}


/*FN**************************************************************************

int classifyLinear(  FILE *out )

Returns: int -- Fail(0) on errors, Succeed(1) otherwise

Purpose: Classify the test data using a linear kernel and write the
         classification result to the prediction file.

Notes: None.
**/

int classifyLinear( FILE *out )
{
  int i;
  double startTime;
  
  printf("Start classifying ...\n");
  startTime = clock()/CLOCKS_PER_SEC;
  for( i = 1; i <= numExample; i++ ) 
    output[i] = wtDotProduct( weight, maxFeature, example[i],
                              nonZeroFeature[i] );
  
  printf( "Classifying time is %f seconds\n",
          clock()/CLOCKS_PER_SEC - startTime );
  printf( "Finish classifying.\n");
  for ( i = 1; i <= numExample; i++ )
    fprintf( out, "%18.17f\n", output[i] - b );
  return 1;
}


/*FN***************************************************************************

int classifyPoly(  FILE *out )

Returns: int -- Fail(0) on errors, Succeed(1) otherwise

Purpose: Classify the test data using a polynomail kernel and write
         the classification results to the prediction file.

Notes: None.
**/


int classifyPoly( FILE *out )
{
  int i, j;
  double startTime;

  printf("Start classifying ...\n");
  startTime = clock()/CLOCKS_PER_SEC;  
  for( i = 1; i <= numExample; i++ ) {
    output[i] = 0;
    for( j = 1; j <= numSv; j++ )
      output[i] += lambda[j] * power(1 + dotProduct(sv[j],svNonZeroFeature[j], 
                   example[i],nonZeroFeature[i]), degree );
  }
  printf( "Classifying time is %f seconds\n",
          clock()/CLOCKS_PER_SEC - startTime );
  printf( "Finish classifying.\n");
  for ( i = 1; i <= numExample; i++ )
    fprintf( out, "%18.17f\n", output[i] - b );    
  return 1;
}



/*FN*************************************************************************
int classifyRbf(  FILE *out )

Returns: int -- Fail(0) on errors, Succeed(1) otherwise

Purpose: Classify the test data using rbf kernel and write the
         classification results to the prediction file.

Notes: None.
**/

int classifyRbf( FILE *out )
{
  int i, j;
  double devSqr;
  double startTime;

  printf("Start classifying ...\n");
  startTime = clock()/CLOCKS_PER_SEC;
  for( i = 1; i <= numExample; i++ ) {
    output[i] = 0;
    for( j = 1; j <= numSv; j++ ) {
      devSqr = dotProduct(sv[j],svNonZeroFeature[j],sv[j],svNonZeroFeature[j]) 
	       - 2*dotProduct(sv[j],svNonZeroFeature[j], example[i],
                              nonZeroFeature[i] )
	       + dotProduct( example[i],nonZeroFeature[i], example[i],
                             nonZeroFeature[i] );
      output[i] += lambda[j] * exp( - devSqr*rbfConstant );
    }
  }
  printf( "Classifying time is %f seconds\n",
          clock()/CLOCKS_PER_SEC - startTime );
  printf( "Finish classifying.\n");                  
  for( i = 1; i <= numExample; i++ )
    fprintf( out, "%18.17f\n", output[i] - b );
  return 1;
}
