/****************************  initialize.c  *******************************

 Purpose: Initialize all data structures for classification by 
          reading information from the model file and the data file.

 Notes: None.
 
 Programmer: Ginny Mak
 Location: Montreal, Canada.
 Last Updated: April 6, 2000. 
**/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "initialize.h"
#include "classify.h"

/********************************* Private Defines and Data Structures *******/

#define MODELTEMP 100 
#define DATATEMP1 10000
#define DATATEMP2 1000


static double C;
static double sigmaSqr;
static maxFeatureRead;

/******************************** Declared Functions  ************************/

static int readString( char *store, char delimiter, FILE *in );
static void skip( char end, FILE *in );
static int initializeModel( int size );
static int initializeData( int dataSize );


/********************************  Private Functions  ************************/

/*FN***************************************************************************

int readString( char *store, char delimiter, FILE *in )
 
Returns: int -- Fail(0) on error, Succeed(1) otherwise.
         a string stored in char *store

Purpose: To read characters from a file until the delimiter is encountered.
         All the characters read except the delimiter are stored in the
         passed character array store.

Notes: The passed file pointer's position changes when the function returns. 
       New line and EOF is not a delimiter. If they are read before the
       delimiter is encountered then the function returns 0.
**/

static int readString( char *store, char delimiter, FILE * in )
{
  char c;
  int i;

  i = 0;
  c = getc( in );
  while (c != delimiter && c != '\n' && c != EOF ) {
    store[i] = c;
    i++;
    c = getc( in );
  }
  if ( c == EOF || c == '\n' )
    return 0;
  else
    store[i] = '\0';
  return 1;
}

/*FN************************************************************************
 
void skip( char end, FILE *in )

Returns: None

Purpose: To skip all the characters read up to the end character.

Note: The passed file pointer's position changes when the function returns.
**/

void skip(char end , FILE * in)
{
  char c;

  while((c = getc( in )) != end)
    ;
}


/*FN***************************************************************************

static int initializeModel( int size )

Returns: int -- Faild(0) on error, Succeed(1) otherwise.

Purpose: To initialize data for readModel function.

Notes: None.
**/

static int initializeModel( int size )
{
  int i;

  lambda = (double *)malloc( (size + 1)* sizeof(double) );
  if ( lambda == NULL) {
    fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
  svNonZeroFeature = (int *) malloc( (size + 1) * sizeof( int ) );
  if (svNonZeroFeature == NULL ) {
    fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
  sv = (FeaturePtr **)malloc( (size + 1) * sizeof(FeaturePtr *));
  if ( sv == NULL ) {
    fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
  for(i = 1; i <= numSv; i++) {
    sv[i] = (FeaturePtr *)malloc((maxFeature) * sizeof(FeaturePtr));
    if ( sv[i] == NULL ) {
      fprintf( stderr, "Memory allocation failed\n");
      return 0;
    }
  }
}

/*FN***************************************************************************
 
static int initializeData( int dataSize ) 

Returns: Int -- Fail(0) on error, Succeed(1) otherwise

Purpose: To initialize data for classification.

Note: None
**/

static int initializeData( int dataSetSize )
{
  example = (FeaturePtr **) malloc( dataSetSize * sizeof(FeaturePtr *) );
  if ( example == NULL ) {
    fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
  output = (double *) malloc( dataSetSize * sizeof(double) );
  if ( output == NULL ) {
      fprintf( stderr, "Memory allocation failed\n");
      return 0;
  }
  target = (int *) malloc( dataSetSize * sizeof(int) ); 
  if ( target == NULL ) {
      fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
  zeroFeatureExample = (int *)malloc( dataSetSize/2 * sizeof(int) );
  if( zeroFeatureExample == NULL ) {
    fprintf( stderr, "Memory allocation failed.\n");
    return 0;
  }                                              
  nonZeroFeature = (int *) malloc( dataSetSize * sizeof(int) );
  if ( nonZeroFeature == NULL ) {
    fprintf( stderr, "Memory allocation failed\n");
    return 0;
  }
}
 


/*******************************  Public Functions   *************************/

  
/*FN***************************************************************************
 
int readModel( FILE *in )

Returns: int -- Fail(0) on error, Succeed(1) otherwise.

Purpose: Read and store informaton read from given model file.

Notes: Assume the model file has been successfully open for reading.
       The model file is a model file resulted from training using
       program smoLearn.
**/

int readModel( FILE *in )
{
  int numNonZeroFeature;	
  int i, j, n;
  char c;
  char temp[MODELTEMP];


  printf("Reading model file...\n");
  c = getc( in );
  /****************************************************
   *                                                  *
   * If c is 0, then the kernel type is linear.       *
   * If c is 1, then the kernel type is polynomial,   *
   * read the degree of polynomial                    *
   * If c is 2, then the kernel type is rbf,          *
   * read the variance.                               *
   *                                                  *
   ****************************************************/
  if( c == '0' )
    kernelType = 0;   
  else if( c == '1' ) {
    kernelType = 1;   
    skip(' ', in);
    readString(temp, ' ', in);
    degree = atoi( temp );
  }
  else if( c == '2' ) {  
    kernelType = 2;
    skip(' ', in);
    readString(temp, ' ', in);
    sigmaSqr = atof( temp );
    rbfConstant = 1/(2*sigmaSqr);
  }
  skip('\n', in);

  /*************************************************
   *                                               *
   * Read number of features of training examples  *
   *                                               *
   *************************************************/
  readString(temp, ' ', in);
  maxFeatureRead = atoi( temp );
  skip('\n', in);
  
  /*************************************************
   *                                               *
   * Read the weight if the kernel type is linear  *
   *                                               * 
   *************************************************/
  if ( kernelType == 0 ) {
    weight = (double *) malloc( (maxFeatureRead+1) * sizeof( double ) );
    if ( weight == NULL ) {
      fprintf( stderr, "Memory allocation failed.\n" );
      return 0;
    }
    i = 0;
    for( i = 0; i <= maxFeatureRead; i++ )
      weight[i] = 0;
    for( i = 1; i <= maxFeatureRead; i++ ) {
      readString( temp, ' ', in );
      weight[i] = atof( temp );
    }
    skip('\n', in);
  }

  /*************************
   *                       *
   * Read the threshold b  *
   *                       *
   *************************/
  readString( temp, ' ', in );
  b = atof( temp );
  skip('\n', in );
  
  /* read C parameter */
  readString( temp, ' ', in );
  C = atof( temp );
  skip('\n', in );

  /******************************************
   *                                        * 
   * Read the number of support vectors     *
   *                                        *
   ******************************************/
  readString( temp, ' ', in );
  numSv = atoi(temp);
  skip( '\n', in );
  if( ! initializeModel( numSv ) )
    return 0;

  /*******************************************************************
   *                                                                 *
   * Read the product of lambda of a support vector with class label *
   *                                                                 *
   *******************************************************************/
  for( i = 1; i <= numSv; i++ ) {
    readString( temp, ' ', in );
    lambda[i] = atof( temp );
    j = 0;
    /***************************************************************
     *                                                             *
     * Read the id/value pairs of the features of a support vector *
     *                                                             * 
     ***************************************************************/
    while( readString( temp, ':', in ) ) {
      sv[i][j] = (FeaturePtr) malloc(sizeof(struct feature));
      if( sv[i][j] == NULL ) {
	fprintf( stderr, "Memory allocation failed\n");
	return 0;
      }
      sv[i][j]->id = atoi( temp );
      if(!readString( temp, ' ', in))
        readString( temp, ' ', in );
      sv[i][j]->value = atof(temp);
      j++;
    }
    svNonZeroFeature[i] = j;
  }
 
  return 1;
}


/*FN**************************************************************************

int readData( FILE *in )

Returns: int -- Fail(0) on error, Succeed(1) otherwise.

Purpose: This function reads the data file to extract  and store the 
         id/value pairs of features of each test data.

Notes: The size of each line of the feature description of data cannot
       be more than 10000 characters long.
       The passed file pointer's position is at EOF if the function
       returns successfully.
**/

int readData( FILE *in )
{
  char temp[DATATEMP1];
  char temp2[DATATEMP2];
  int numFeature, maxFeature;
  int exampleIndex, featureIndex;
  int i,j,dataSetSize;
  char c;
  int zeroFeatureNumber = 0;

  dataSetSize = 0;
  /***********************************
   *                                 *
   * Estimate the number of data.    *
   *                                 *
   ***********************************/        
  while( (c = getc( in )) != EOF ) {
    if( c == '\n' )
      dataSetSize++;
  }
  dataSetSize++;
  rewind ( in );

  if( !initializeData( dataSetSize ) )
    return 0;
 
  /*************************************************************
   *                                                           *
   * NonZeroFeature[0] stores the number of non-zero features  *
   * of the weight vector.                                     *
   *                                                           *
   *************************************************************/
  nonZeroFeature[0] = maxFeature;
  
  numExample = 0; exampleIndex = 0;
  printf("Reading test data file ...\n");
  while( ( c = getc(in) ) != EOF ) {
    while( c == '#' || c == '\n' ) {
      /***************************************
       *                                     *
       * Ignore comment and blank line       *
       *                                     *
       ***************************************/
      if( c == '#' ) {
	while( ( c = getc(in) ) != '\n' )
	  ;
      }
      if( c == '\n' ) 
        c = getc( in ); 
    }
    if( c == EOF )
      break;
    
    /******************************
     *                            *
     * Read one line of input     *
     *                            *
     ******************************/
    else {
      exampleIndex++;
      i = 0; numFeature = 0;
      temp[i] = c; i++;
      while( ( c = getc(in) ) != '\n' ) {
	temp[i] = c; i++;
	if( c == ':' )
	  numFeature++;
      }
      temp[i] = '\0';
      numExample++;
      nonZeroFeature[exampleIndex] = numFeature;

      if ( numFeature != 0 ) {
	/****************************************************************
	 *                                                              *
	 * Allocate memory for id/value pairs of the test data          *
	 *                                                              *
	 ****************************************************************/
	example[exampleIndex] = (FeaturePtr *)malloc( numFeature *
                               sizeof( FeaturePtr ) );
	if ( example[exampleIndex] == NULL ) {
	  fprintf( stderr, "Memory allocation failed\n");
	  return 0;
	}
	for (j = 0; j < numFeature; j++) {
	  example[exampleIndex][j] = (FeaturePtr)
                                    malloc (sizeof(struct feature));
	  if ( example[exampleIndex][j] == NULL ) {
	    fprintf( stderr, "Memory allocation failed\n");
	    return 0;
	  }
	}
      }
      else {
	 example[exampleIndex] = (FeaturePtr *)
                                  malloc( sizeof( FeaturePtr ) );
         if ( example[exampleIndex] == NULL ) {
	   fprintf( stderr, "Memory allocation failed\n");
	   return 0;
	 }
	 example[exampleIndex][0] = (FeaturePtr)
                                     malloc( sizeof(struct feature) );
	 if ( example[exampleIndex][0] == NULL ) {       
       	    fprintf( stderr, "Memory allocation failed\n");
            return 0;
	 }
	 example[exampleIndex][0]->id = 1;
	 example[exampleIndex][0]->value = 0;
	 nonZeroFeature[exampleIndex] = 0;
	 zeroFeatureExample[zeroFeatureNumber] = exampleIndex;
	 zeroFeatureNumber++;
      }                                        
	
      /**********************************
       *                                *
       * Extract class of example       *
       *                                *
       **********************************/
      i = 0;
      while( temp[i] != ' ' ) {
	temp2[i] = temp[i]; i++;
      }
      temp2[i] = '\0';
      target[exampleIndex] = atoi(temp2);
      i++;

      if ( numFeature != 0 ) {
	/***********************************************************
	 *                                                         *
	 * Extract id/value pairs of the features of the test data *
	 *                                                         *
	 ***********************************************************/
	featureIndex = 0;
	while( temp[i] != '\0' ) {
	  j = 0;
	  while( temp[i] != ':' ) {
	    temp2[j] = temp[i]; i++; j++;
	  }
	  temp2[j] = '\0';
	  example[exampleIndex][featureIndex]->id = atoi(temp2);
	  j = 0; i++;
	  while( temp[i] != ' ' && temp[i] != '\0' ) {
	    temp2[j] = temp[i];
	    i++; j++;
	  }
	  temp2[j] = '\0';

	  if( atof(temp2) != 0 ) {
	    example[exampleIndex][featureIndex]->value = atof(temp2);
	    featureIndex++; 

	  }
	} /* end while */

	nonZeroFeature[exampleIndex] = featureIndex;
	if( example[exampleIndex][featureIndex - 1]->id > maxFeature )
	  maxFeature = example[exampleIndex][featureIndex - 1]->id;
      }
    } /* end else */
  } /* end of while not EOF */

  /**********************************************************************
   *                                                                    *
   * If the maxFeature read from model file is less than the number of  *
   * features the test data has, then extend the weight vector.         *
   *                                                                    *
   **********************************************************************/
  if ( maxFeatureRead < maxFeature ) {
    weight = realloc( weight, maxFeature+1 );
    for ( i = maxFeatureRead+1; i <= maxFeature; i++ )
      weight[i] = 0;
  }
    
  return 1;
}

